from functools import partial
import inspect
from typing import Any, Callable

AnyCallable = Callable[..., Any]


class SelfInspector:
    """
    This class provides a utility method `resolve_callable` that can be used to
    determine whether a callable belongs to the instance itself and returns another
    callable, with the instance bound to it.
    """

    def resolve_callable(self, callable: AnyCallable) -> AnyCallable:
        if self.is_instance_method(callable):
            return partial(callable, self)
        elif self.is_class_method(callable):
            return partial(callable.__func__, self.__class__)
        elif self.is_static_method(callable):
            return callable.__func__

        return callable

    @classmethod
    def is_instance_method(cls, callable: AnyCallable) -> bool:
        functions = [
            function for _, function in inspect.getmembers(cls, inspect.isfunction)
        ]
        return callable in functions

    @classmethod
    def is_class_method(cls, callable: AnyCallable) -> bool:
        methods = [
            method.__func__ for _, method in inspect.getmembers(cls, inspect.ismethod)
        ]
        return isinstance(callable, classmethod) and callable.__func__ in methods

    @classmethod
    def is_static_method(cls, callable: AnyCallable) -> bool:
        functions = [
            function for _, function in inspect.getmembers(cls, inspect.isfunction)
        ]
        return isinstance(callable, staticmethod) and callable.__func__ in functions



import inspect
from types import UnionType
from typing import Optional, Union, get_type_hints
from pydantic import BaseModel


def get_required_fields(model_cls: type[BaseModel]) -> list[str]:
    required_fields = list()

    type_hints = get_type_hints(model_cls)
    for field, field_type in type_hints.items():
        # For Python 3.10 and later: Check if field_type is UnionType and contains None
        if isinstance(field_type, UnionType):
            if type(None) not in field_type.__args__:
                required_fields.append(field)
        # For earlier versions: Check if field_type is a Union and if NoneType is one of the Union types
        elif getattr(field_type, "__origin__", None) is Union:
            if type(None) not in getattr(field_type, "__args__", ()):
                required_fields.append(field)
        # Check if the type hint is explicitly Optional but not a Union (for completeness)
        elif not (
            inspect.isclass(field_type) and issubclass(field_type, type(Optional))
        ):
            required_fields.append(field)

    return required_fields



from datetime import datetime
from pydantic import BaseModel
from pytest import raises
from py_transmuter.pydantic.mapper import BaseModelMapper


def test_map_with_field_name():
    class A(BaseModel):
        id: int

    class B(BaseModel):
        id: int

    class ABMapper(BaseModelMapper[B, A]):
        mapping = {"id": "id"}

    assert ABMapper().map(A(id=1)) == B(id=1)


def test_map_with_lambda():
    class A(BaseModel):
        id: int

    class B(BaseModel):
        id: str

    class ABMapper(BaseModelMapper[B, A]):
        mapping = {"id": lambda data: str(data.id)}

    assert ABMapper().map(A(id=1)) == B(id="1")


def test_map_with_function():
    class A(BaseModel):
        id: int

    class B(BaseModel):
        id: str

    def map_id(data: A) -> str:
        return str(data.id * 10)

    class ABMapper(BaseModelMapper[B, A]):
        mapping = {"id": map_id}

    assert ABMapper().map(A(id=1)) == B(id="10")


def test_map_with_context_and_self_inspection():
    class A(BaseModel):
        id: int

    class B(BaseModel):
        id: str

    class ABMapper(BaseModelMapper[B, A]):
        def map_id(self, data: A) -> str:
            return str(data.id * self.context["factor"])

        mapping = {"id": map_id}

    assert ABMapper(context={"factor": 2}).map(A(id=1)) == B(id="2")


def test_map_with_class_method():
    class A(BaseModel):
        id: int

    class B(BaseModel):
        id: int

    class ABMapper(BaseModelMapper[B, A]):
        FACTOR = 10

        @classmethod
        def map_id(cls, data: A) -> int:
            return data.id * cls.FACTOR

        mapping = {"id": map_id}

    assert ABMapper().map(A(id=1)) == B(id=10)


def test_map_with_tuple_accessor():
    class A(BaseModel):
        id: int

    class B(BaseModel):
        id: int

    class ABMapper(BaseModelMapper[B, A]):
        mapping = {"id": ("id", lambda value: value * 10)}

    assert ABMapper().map(A(id=1)) == B(id=10)


def test_map_list():
    class A(BaseModel):
        id: int

    class B(BaseModel):
        id: int

    class ABMapper(BaseModelMapper[B, A]):
        mapping = {"id": "id"}

    assert ABMapper().map_list([A(id=1), A(id=2)]) == [B(id=1), B(id=2)]


def test_map_without_optional_fields():
    class A(BaseModel):
        id: int

    class B(BaseModel):
        id: int
        name: str | None = None

    class ABMapper(BaseModelMapper[B, A]):
        mapping = {"id": "id"}

    assert ABMapper().map(A(id=1)) == B(id=1)


def test_map_missing_required_field_fails():
    class A(BaseModel):
        id: int

    class B(BaseModel):
        id: int
        name: str

    class ABMapper(BaseModelMapper[B, A]):
        mapping = {"id": "id"}

    with raises(ValueError):
        ABMapper()


def test_map_extra_field_fails():
    class A(BaseModel):
        id: int

    class B(BaseModel):
        id: int

    class ABMapper(BaseModelMapper[B, A]):
        mapping = {"id": "id", "name": "name"}

    with raises(ValueError):
        ABMapper()


def test_map_with_optional_field():
    class A(BaseModel):
        id: int

    class B(BaseModel):
        id: int
        name: str | None = None

    class ABMapper(BaseModelMapper[B, A]):
        mapping = {"id": "id", "name": lambda data: "Rodrigo"}

    assert ABMapper().map(A(id=1)) == B(id=1, name="Rodrigo")


def test_map_complex_models():
    class WeatherDataBackend(BaseModel):
        temperature_fahrenheit: float
        humidity_proportion: float
        timestamp: datetime
        
        country: str | None = None

    class WeatherDataUI(BaseModel):
        temperature_celsius: float
        humidity_percentage: int
        date: str
        time: str

        language: str

    class WeatherDataMapper(BaseModelMapper[WeatherDataBackend, WeatherDataUI]):
        @staticmethod
        def celsius_to_fahrenheit(celsius: float) -> float:
            return (celsius * 9 / 5) + 32

        @staticmethod
        def parse_date_time_to_object(data: WeatherDataUI) -> datetime:
            return datetime.strptime(
                f"{data.date} {data.time}", "%Y-%m-%d %H:%M"
            )

        mapping = {
            "temperature_fahrenheit": ("temperature_celsius", celsius_to_fahrenheit),
            "humidity_proportion": ("humidity_percentage", lambda humidity: float(humidity) / 100),
            "timestamp": parse_date_time_to_object,
        }

    to_map = WeatherDataUI(
        temperature_celsius=10,
        humidity_percentage=50,
        date="2021-01-01", 
        time="12:00",
        language="en",
    )

    mapped = WeatherDataMapper().map(to_map)
    assert mapped == WeatherDataBackend(
        temperature_fahrenheit=50.0,
        humidity_proportion=0.5,
        timestamp=datetime(2021, 1, 1, 12)
    )



from datetime import date, datetime, timedelta
from statistics import mean
from pydantic import BaseModel
from pytest import raises
from py_transmuter.pydantic.aggregator import BaseModelAggregator


def test_aggregator_field_direct_mappings():
    class A(BaseModel):
        id: int

    class B(BaseModel):
        ids: list[int]

    class ABAggregator(BaseModelAggregator[B, A]):
        mappings = {"ids": "id"}

    assert ABAggregator().aggregate([A(id=1), A(id=2), A(id=3)]) == [B(ids=[1, 2, 3])]


def test_aggregator_tuple_direct_mappings():
    class A(BaseModel):
        id: int

    class B(BaseModel):
        ids: list[str]

    class ABAggregator(BaseModelAggregator[B, A]):
        mappings = {"ids": ("id", lambda id: str(id))}

    assert ABAggregator().aggregate([A(id=1), A(id=2), A(id=3)]) == [
        B(ids=["1", "2", "3"])
    ]


def test_aggregator_model_callable_direct_mappings():
    class A(BaseModel):
        first_name: str
        last_name: str

    class B(BaseModel):
        full_names: list[str]

    def make_full_name(data: A) -> str:
        return f"{data.first_name} {data.last_name}"

    class ABAggregator(BaseModelAggregator[B, A]):
        mappings = {"full_names": make_full_name}

    assert ABAggregator().aggregate(
        [
            A(first_name="Rodrigo", last_name="De Rosa"),
            A(first_name="Marcelo", last_name="Gallardo"),
            A(first_name="Super", last_name="Man"),
        ]
    ) == [B(full_names=["Rodrigo De Rosa", "Marcelo Gallardo", "Super Man"])]


def test_aggregator_field_direct_mappings_with_grouping():
    class Child(BaseModel):
        parent: str

        first_name: str
        last_name: str

    class Parent(BaseModel):
        children_names: list[str]

    def make_full_name(data: Child) -> str:
        return f"{data.first_name} {data.last_name}"

    class ChildParentAggregator(BaseModelAggregator[Parent, Child]):
        group_by = ("parent",)

        mappings = {"children_names": make_full_name}

    assert ChildParentAggregator().aggregate(
        [
            Child(parent="Patricio", first_name="Rodrigo", last_name="De Rosa"),
            Child(parent="God Himself", first_name="Marcelo", last_name="Gallardo"),
            Child(parent="God Himself", first_name="Super", last_name="Man"),
        ]
    ) == [
        Parent(children_names=["Rodrigo De Rosa"]),
        Parent(children_names=["Marcelo Gallardo", "Super Man"]),
    ]


def test_aggregator_group_by_with_extractor():
    class Child(BaseModel):
        parent: str

        first_name: str
        last_name: str

    class Parent(BaseModel):
        children_names: list[str]

    class ChildParentAggregator(BaseModelAggregator[Parent, Child]):
        @staticmethod
        def parent_full_name(data: Child) -> str:
            return f"{data.parent} {data.last_name}"

        group_by = (parent_full_name,)

        mappings = {"children_names": "first_name"}

    assert ChildParentAggregator().aggregate(
        [
            Child(parent="Tom", first_name="Paul", last_name="Doe"),
            Child(parent="Tom", first_name="George", last_name="Stones"),
            Child(parent="Tom", first_name="Carl", last_name="Doe"),
        ]
    ) == [Parent(children_names=["Paul", "Carl"]), Parent(children_names=["George"])]


def test_aggregator_with_grouping_by_field_and_extractor():
    class Measurement(BaseModel):
        machine: str
        timestamp: datetime
        value: float

    class DayValues(BaseModel):
        intervals: list[datetime]
        values: list[float]

    class DailyMeasurementAggregator(BaseModelAggregator[DayValues, Measurement]):
        @staticmethod
        def timestamp_date(data: Measurement) -> date:
            return data.timestamp.date()

        group_by = ("machine", timestamp_date)

        mappings = {"intervals": "timestamp", "values": "value"}

    aggregator = DailyMeasurementAggregator()

    day_values = aggregator.aggregate(
        [
            Measurement(
                machine="A",
                timestamp=datetime(2024, 2, 16) + timedelta(hours=12 * i),
                value=100,
            )
            for i in range(4)
        ]
    )

    assert day_values == [
        DayValues(
            intervals=[datetime(2024, 2, 16), datetime(2024, 2, 16, 12)],
            values=[100, 100],
        ),
        DayValues(
            intervals=[datetime(2024, 2, 17), datetime(2024, 2, 17, 12)],
            values=[100, 100],
        ),
    ]


def test_aggregator_with_sort_by_field():
    class A(BaseModel):
        id: int

    class B(BaseModel):
        ids: list[int]

    class ABAggregator(BaseModelAggregator[B, A]):
        sort_by = ("id",)

        mappings = {"ids": "id"}

    assert ABAggregator().aggregate([A(id=3), A(id=1), A(id=2)]) == [B(ids=[1, 2, 3])]


def test_aggregator_with_sort_by_extractor():
    class Child(BaseModel):
        first_name: str
        last_name: str

    class Parent(BaseModel):
        children_names: list[str]

    def make_full_name(data: Child) -> str:
        return f"{data.first_name} {data.last_name}"

    class ChildParentAggregator(BaseModelAggregator[Parent, Child]):
        sort_by = (make_full_name,)

        mappings = {"children_names": make_full_name}

    assert ChildParentAggregator().aggregate(
        [
            Child(first_name="Paul", last_name="Smith"),
            Child(first_name="Paul", last_name="Holmes"),
        ]
    ) == [Parent(children_names=["Paul Holmes", "Paul Smith"])]


def test_aggregator_with_sort_by_extractor_and_field():
    class Child(BaseModel):
        first_name: str
        last_name: str

        age: int

    class Parent(BaseModel):
        children_names: list[str]
        ages: list[int]

    def make_full_name(data: Child) -> str:
        return f"{data.first_name} {data.last_name}"

    class ChildParentAggregator(BaseModelAggregator[Parent, Child]):
        sort_by = (make_full_name, "age")

        mappings = {"children_names": make_full_name, "ages": "age"}

    assert ChildParentAggregator().aggregate(
        [
            Child(first_name="Paul", last_name="Smith", age=17),
            Child(first_name="Paul", last_name="Holmes", age=23),
            Child(first_name="Paul", last_name="Smith", age=5),
        ]
    ) == [
        Parent(
            children_names=["Paul Holmes", "Paul Smith", "Paul Smith"], ages=[23, 5, 17]
        )
    ]


def test_aggregator_with_field_mapper_aggregations():
    class Vertical(BaseModel):
        value: float

    class Horizontal(BaseModel):
        values: list[float]

    def aggregate_values(values: list[float]) -> float:
        average = sum(values) / len(values)
        return [value / average for value in values]

    class VerticalToHorizontalAggregator(BaseModelAggregator[Horizontal, Vertical]):
        aggregations = {
            "values": ("value", aggregate_values),
        }

    assert VerticalToHorizontalAggregator().aggregate(
        [Vertical(value=6), Vertical(value=4), Vertical(value=2)]
    ) == [Horizontal(values=[1.5, 1, 0.5])]

    
def test_aggregator_with_extractor_aggregations():
    class Coordinates(BaseModel):
        latitude: float
        longitude: float

    class Turbine(BaseModel):
        latitude: float
        longitude: float

    class Park(BaseModel):
        location: Coordinates

    def average_coordinates(turbines: list[Turbine]) -> Coordinates:
        return Coordinates(
            latitude=mean(turbine.latitude for turbine in turbines),
            longitude=mean(turbine.longitude for turbine in turbines),
        )

    class TurbineParkAggregator(BaseModelAggregator[Park, Turbine]):
        aggregations = {
            "location": average_coordinates,
        }

    assert TurbineParkAggregator().aggregate(
        [
            Turbine(latitude=10, longitude=5), 
            Turbine(latitude=15, longitude=10),
            Turbine(latitude=20, longitude=0),
        ]
    ) == [Park(location=Coordinates(latitude=15, longitude=5))]


def test_aggregator_with_mappings_and_aggregations():
    class A(BaseModel):
        parent_id: int
        value: float

    class B(BaseModel):
        id: int
        values: list[float]

    class ABAggregator(BaseModelAggregator[B, A]):
        group_by = ("parent_id",)

        mappings = {"values": "value"}
        aggregations = {"id": ("parent_id", lambda ids: ids[0])}

    assert ABAggregator().aggregate(
        [A(parent_id=1, value=10), A(parent_id=1, value=20), A(parent_id=2, value=30)]
    ) == [B(id=1, values=[10, 20]), B(id=2, values=[30])]


def test_aggregator_with_overlapping_mappings_and_aggregations_fails_to_create():
    class A(BaseModel):
        parent_id: int

    class B(BaseModel):
        id: int

    class ABAggregator(BaseModelAggregator[B, A]):
        mappings = {"id": "parent_id"}
        aggregations = {"id": ("parent_id", lambda ids: ids[0])}


    with raises(ValueError):
        ABAggregator()


def test_aggregator_without_mappings_or_aggregations():
    class A(BaseModel):
        id: int

    class B(BaseModel):
        id: int

    class ABAggregator(BaseModelAggregator[B, A]):
        pass

    with raises(ValueError):
        ABAggregator()


def test_aggregator_missing_required_fields():
    class A(BaseModel):
        id: int
        value: float

    class B(BaseModel):
        id: int
        values: list[float]

    class ABAggregator(BaseModelAggregator[B, A]):
        aggregations = {"values": "value"}

    with raises(ValueError):
        ABAggregator()


def test_aggregator_with_extra_fields():
    class A(BaseModel):
        id: int
        value: float

    class B(BaseModel):
        id: int

    class ABAggregator(BaseModelAggregator[B, A]):
        aggregations = {"id": "id", "values": "value"}

    with raises(ValueError):
        ABAggregator()